"""
Turbomole Writer
================

"""


class TurbomoleWriter:
    """
    A writer class for ``Turbomole`` files.

    Examples
    --------
    *Writing to a File with a Unit Cell*

    This writer can write to a file with the unit
    cell included for periodic molecules. Note that this always assumes
    P1 space group.

    .. testcode:: writing-to-a-file-with-a-unit-cell

        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])
        topology_graph = stk.cof.PeriodicHoneycomb(
            building_blocks=(bb1, bb2),
            lattice_size=(3, 3, 1),
        )
        construction_result = topology_graph.construct()
        cof = stk.ConstructedMolecule.init_from_construction_result(
            construction_result=construction_result,
        )
        writer = stk.TurbomoleWriter()
        writer.write(
            molecule=cof,
            path='cof.coord',
            periodic_info=construction_result.get_periodic_info(),
        )

    .. testcode:: writing-to-a-file-with-a-unit-cell
        :hide:

        import os

        assert os.path.exists('cof.coord')

    .. testcleanup:: writing-to-a-file-with-a-unit-cell

        os.remove('cof.coord')

    """

    def _write_content(self, molecule, atom_ids, periodic_info=None):

        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )

        content = []
        if periodic_info is not None:
            # Input unit cell information.
            a = periodic_info.get_a()
            b = periodic_info.get_b()
            c = periodic_info.get_c()
            alpha = periodic_info.get_alpha()
            beta = periodic_info.get_beta()
            gamma = periodic_info.get_gamma()
            content.append(
                '$periodic 3\n'
                '$cell angs\n'
                f' {a:>8.3f} {b:>8.3f} {c:>8.3f} '
                f'{alpha:>6.2f} {beta:>6.2f} {gamma:>6.2f}\n'
            )

        coords = molecule.get_position_matrix()
        content.append('$coord angs\n')
        for atom_id in atom_ids:
            atom, = molecule.get_atoms(atom_ids=atom_id)
            element = atom.__class__.__name__
            x, y, z = (i for i in coords[atom_id])
            content.append(
                f' {round(x, 4)} {round(y, 4)} {round(z, 4)} '
                f'{element}\n'
            )

        content.append('$end\n')

        return content

    def to_string(
        self,
        molecule,
        atom_ids=None,
        periodic_info=None
    ):
        """
        Get a ``Turbomole`` file format string of `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to ``Turbomole`` format.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        periodic_info : :class:`.PeriodicInfo`
            Information about the periodic cell.

        Returns
        -------
        :class:`string`
            The content of a `.coord` file.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        return ''.join(content)

    def write(self, molecule, path, atom_ids=None, periodic_info=None):
        """
        Write `molecule` to ``Turbomole`` file format.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to ``Turbomole`` format.

        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        periodic_info : :class:`.PeriodicInfo`
            Information about the periodic cell.

        Returns
        -------
        None : :class:`NoneType`
            A file is written.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        with open(path, 'w') as f:
            f.write(''.join(content))



"""
Random Crosser
==============

"""


import numpy as np


class RandomCrosser:
    """
    Use some other crosser at random.

    Examples
    --------
    *Use One of Several Crossers at Random*

    .. testcode:: use-one-of-several-crossers-at-random

        import stk

        def get_functional_group_type(building_block):
            fg, = building_block.get_functional_groups(0)
            return type(fg)

        def get_num_functional_groups(building_block):
           return building_block.get_num_functional_groups()

        crosser = stk.RandomCrosser(
            crossers=(
                stk.GeneticRecombination(get_functional_group_type),
                stk.GeneticRecombination(get_num_functional_groups),
            ),
            random_seed=12,
        )

        record1 = stk.MoleculeRecord(
            topology_graph=stk.cage.FourPlusSix(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='O=CC(C=O)CC=O',
                        functional_groups=[stk.AldehydeFactory()],
                    ),
                    stk.BuildingBlock(
                        smiles='NCCN',
                        functional_groups=[stk.PrimaryAminoFactory()],
                    ),
                ),
            ),
        )
        record2 = stk.MoleculeRecord(
            topology_graph=stk.cage.FourPlusSix(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='O=CNC(C=O)CC=O',
                        functional_groups=[stk.AldehydeFactory()],
                    ),
                    stk.BuildingBlock(
                        smiles='NCNCN',
                        functional_groups=[stk.PrimaryAminoFactory()],
                    ),
                ),
            ),
        )

        # Use one of the component crossers at random.
        crossover_records1 = tuple(crosser.cross((record1, record2)))
        # A different crosser may get selected at random the second,
        # third, etc, time.
        crossover_records2 = tuple(crosser.cross((record1, record2)))

    .. testcode:: use-one-of-several-crossers-at-random
        :hide:

        _expected_results = (
            stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CNC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCNCN',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                    ),
                ),
            ),
            stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCCN',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                    ),
                ),
            ),
            stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CNC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCCN',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                    ),
                ),
            ),
            stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCNCN',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                    ),
                ),
            ),
        )

        def _get_smiles(item):
            if isinstance(item, stk.ConstructedMolecule):
               return stk.Smiles().get_key(item)
            return stk.Smiles().get_key(
                molecule=item.get_molecule_record().get_molecule(),
             )

        _expected_smiles = set(map(_get_smiles, _expected_results))

        _crossover_smiles1 = set(map(_get_smiles, crossover_records1))
        _crossover_smiles2 = set(map(_get_smiles, crossover_records2))
        assert _expected_smiles == _crossover_smiles1
        assert _expected_smiles == _crossover_smiles2

    """

    def __init__(self, crossers, weights=None, random_seed=None):
        """
        Initialize a :class:`.RandomCrosser` instance.

        Parameters
        ----------
        crossers : :class:`tuple`
            Holds instances which have :meth:`.cross` method. The
            :meth:`.cross` method must return an instance of
            :class:`.CrossoverRecord`.

        weights : :class:`tuple` of :class:`float`, optional
            For each crosser, the probability that it will be chosen
            whenever :meth:`.cross` is called.
            If ``None`` all `crossers` will have equal chance of being
            selected.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._crossers = crossers
        self._weights = weights
        self._generator = np.random.RandomState(random_seed)

    def cross(self, records):
        """
        Cross `records`.

        Parameters
        ----------
        records : :class:`iterable` of :class:`.MoleculeRecord`
            The molecule records on which a crossover operation is
            performed.

        Yields
        -------
        :class:`.CrossoverRecord`
            A record of a crossover operation.

        """

        crosser = self._generator.choice(
            a=self._crossers,
            p=self._weights,
        )
        return crosser.cross(records)



from .mae import _with_structure_from_mae  # noqa
from .mdl_mol import _with_structure_from_mol  # noqa
from .pdb import _with_structure_from_pdb  # noqa
from .turbomole import _with_structure_from_turbomole  # noqa
from .xyz import _with_structure_from_xyz  # noqa



"""
Stochastic Universal Sampling
=============================

"""

import numpy as np

from stk.molecular import Inchi

from .selector import Selector


class StochasticUniversalSampling(Selector):
    """
    Yields batches of molecules through stochastic universal sampling.

    Stochastic universal sampling lays out batches along a line, with
    each batch taking up length proportional to its fitness. It
    then creates a set of evenly spaced pointers to different points
    on the line, each of which is occupied by a batch. Batches which
    are pointed to are yielded.

    This approach means weaker members of the population
    are given a greater chance to be chosen than in
    :class:`.Roulette` selection [#]_.

    References
    ----------
    .. [#] https://en.wikipedia.org/wiki/Stochastic_universal_sampling

    Examples
    --------
    *Yielding Single Molecule Batches*

    Yielding molecules one at a time. For example, if molecules need
    to be selected for mutation or the next generation.

    .. testcode:: yielding-single-molecule-batches

        import stk

        # Make the selector.
        stochastic_sampling = stk.StochasticUniversalSampling(5)

        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(100)
        )

        # Select the molecules.
        for selected, in stochastic_sampling.select(population):
            # Do stuff with each selected molecule.
            pass

    """

    def __init__(
        self,
        num_batches=None,
        batch_size=1,
        duplicate_molecules=True,
        duplicate_batches=True,
        key_maker=Inchi(),
        fitness_modifier=None,
        random_seed=None,
    ):
        """
        Initialize a :class:`.StochasticUniversalSampling` instance.

        Parameters
        ----------
        num_batches : :class:`int`, optional
            The number of batches to yield. If ``None`` then yielding
            will continue forever or until the generator is exhausted,
            whichever comes first.

        batch_size : :class:`int`, optional
            The number of molecules yielded at once.

        duplicate_molecules : :class:`bool`, optional
            If ``True`` the same molecule can be yielded in more than
            one batch.

        duplicate_batches : :class:`bool`, optional
            If ``True`` the same batch can be yielded more than once.

        key_maker : :class:`.MoleculeKeyMaker`, optional
            Used to get the keys of molecules. If two molecules have
            the same key, they are considered duplicates.

        fitness_modifier : :class:`callable`, optional
            Takes the `population` on which :meth:`.select` is called
            and returns a :class:`dict`, which maps records in the
            `population` to the fitness values the :class:`.Selector`
            should use. If ``None``, the regular fitness values of the
            records are used.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        if fitness_modifier is None:
            fitness_modifier = self._get_fitness_values

        if num_batches is None:
            num_batches = float('inf')

        self._generator = np.random.RandomState(random_seed)
        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = num_batches
        self._batch_size = batch_size
        super().__init__(
            key_maker=key_maker,
            fitness_modifier=fitness_modifier,
        )

    def _select_from_batches(self, batches, yielded_batches):
        batches = sorted(batches, reverse=True)

        # SUS may need to run multiple rounds if duplicate_molecules or
        # duplicate_batches is False. This is because in each round
        # you can generate multiple pointers to the same batch or to
        # batches sharing molecules. If this happens the lower fitness
        # batch will not be yielded. Instead a second round of SUS will
        # occur with any ineligible batches removed and a reduced
        # number of pointers, to account for batches yielded in the
        # previous rounds. This will repeat until the desired number
        # of batches has been yielded, or there are no more valid
        # batches.
        while (
            batches
            and yielded_batches.get_num() < self._num_batches
        ):
            yield from self._select_with_stochastic_universal_sampling(
                batches=batches,
                yielded_batches=yielded_batches,
            )

            if yielded_batches.get_num() < self._num_batches:
                if not self._duplicate_molecules:
                    batches = filter(
                        yielded_batches.has_no_yielded_molecules,
                        batches,
                    )
                if not self._duplicate_batches:
                    batches = filter(
                        yielded_batches.is_unyielded_batch,
                        batches,
                    )
                if (
                    not self._duplicate_molecules
                    or not self._duplicate_batches
                ):
                    batches = tuple(batches)

    def _select_with_stochastic_universal_sampling(
        self,
        batches,
        yielded_batches,
    ):

        total = sum(batch.get_fitness_value() for batch in batches)
        batch_positions = []
        batch_position = 0
        for batch in batches:
            batch_position += batch.get_fitness_value()/total
            batch_positions.append(batch_position)

        num_batches = min(
            self._num_batches - yielded_batches.get_num(),
            len(batches)
        )
        pointer_distance = 1/num_batches
        pointers = []
        pointer = self._generator.uniform(0, pointer_distance)
        for i in range(num_batches):
            pointers.append(pointer)
            pointer += pointer_distance

        batch_index = 0
        for pointer in pointers:
            while pointer > batch_positions[batch_index]:
                batch_index += 1

            batch = batches[batch_index]

            if (
                not self._duplicate_molecules
                and yielded_batches.has_yielded_molecules(batch)
            ):
                continue

            if (
                not self._duplicate_batches
                and yielded_batches.is_yielded_batch(batch)
            ):
                continue

            yield batch



"""
Paddlewheel
===========

"""

from ...topology_graph import Edge
from ..metal_complex import MetalComplex
from ..vertices import BiDentateLigandVertex, MetalVertex


class Paddlewheel(MetalComplex):
    """
    Represents a metal complex topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Cu+2]',
            functional_groups=(
                stk.SingleAtom(stk.Cu(0, charge=2))
                for i in range(4)
            ),
            position_matrix=([0, 0, 0], ),
        )
        bb2 = stk.BuildingBlock(
            smiles='O=C(O)c1ccc(Br)cc1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#8]~[#1]',
                    bonders=(1, ),
                    deleters=(2, ),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#8X1]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )
        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.Paddlewheel(
                metals=bb1,
                ligands=bb2,
                reaction_factory=stk.DativeReactionFactory(
                    stk.GenericReactionFactory(
                        bond_orders={
                            frozenset({
                                stk.GenericFunctionalGroup,
                                stk.SingleAtom,
                            }): 9,
                        },
                    ),
                ),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    complex.get_atoms(),
                    complex.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in complex.get_bonds()
            ),
        )

    Metal building blocks with at least four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1)
        | ligands: (0, 1, 2, 3)

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (
        MetalVertex(0, [0, 1, 0]),
        MetalVertex(1, [0, -1, 0]),
    )
    _ligand_vertex_prototypes = (
        BiDentateLigandVertex(2, [2, 0, 0]),
        BiDentateLigandVertex(3, [0, 0, 2]),
        BiDentateLigandVertex(4, [-2, 0, 0]),
        BiDentateLigandVertex(5, [0, 0, -2]),
    )

    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0.1, 0.5, 0],
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[1],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0.1, -0.5, 0],
        ),

        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
            position=[0, 0.5, 0.1],
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[1],
            vertex2=_ligand_vertex_prototypes[1],
            position=[0, -0.5, 0.1],
        ),

        Edge(
            id=4,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[2],
            position=[-0.1, 0.5, 0],
        ),
        Edge(
            id=5,
            vertex1=_metal_vertex_prototypes[1],
            vertex2=_ligand_vertex_prototypes[2],
            position=[-0.1, -0.5, 0],
        ),

        Edge(
            id=6,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[3],
            position=[0, 0.5, -0.1],
        ),
        Edge(
            id=7,
            vertex1=_metal_vertex_prototypes[1],
            vertex2=_ligand_vertex_prototypes[3],
            position=[0, -0.5, -0.1],
        ),
    )



from .collapser import *  # noqa
from .mchammer import *  # noqa
from .null import *  # noqa
from .optimizer import *  # noqa
from .periodic_collapser import *  # noqa
from .spinner import *  # noqa
from .utilities import *  # noqa



"""
Alcohol
=======

"""

from .generic_functional_group import GenericFunctionalGroup


class Alcohol(GenericFunctionalGroup):
    """
    Represents an alcohol functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][oxygen][hydrogen]``.

    """

    def __init__(
        self,
        oxygen,
        hydrogen,
        atom,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initialize a :class:`.Alcohol` instance.

        Parameters
        ----------
        oxygen : :class:`.O`
            The oxygen atom.

        hydrogen : :class:`.H`
            The hydrogen atom.

        atom : :class:`.Atom`
            The atom to which the alcohol is attached.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._oxygen = oxygen
        self._hydrogen = hydrogen
        self._atom = atom
        atoms = (oxygen, hydrogen, atom)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_oxygen(self):
        """
        Get the oxygen atom.

        Returns
        -------
        :class:`.O`
            The oxygen atom.

        """

        return self._oxygen

    def get_hydrogen(self):
        """
        Get the hydrogen atom.

        Returns
        -------
        :class:`.H`
            The hydrogen atom.

        """

        return self._hydrogen

    def get_atom(self):
        """
        Get the atom to which the functional group is attached.

        Returns
        -------
        :class:`.Atom`
            The atom to which the functional group is attached.

        """

        return self._atom

    def clone(self):
        clone = super().clone()
        clone._oxygen = self._oxygen
        clone._hydrogen = self._hydrogen
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._oxygen = atom_map.get(
            self._oxygen.get_id(),
            self._oxygen,
        )
        clone._hydrogen = atom_map.get(
            self._hydrogen.get_id(),
            self._hydrogen,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self._oxygen}, {self._hydrogen}, {self._atom}, '
            f'bonders={self._bonders}, deleters={self._deleters})'
        )



"""
XYZ Writer
==========

"""


class XyzWriter:
    """
    A writer class for ``.xyz`` files.

    Examples
    --------
    *Writing to a File*

    .. testcode:: writing-to-a-file

        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])

        writer = stk.XyzWriter()
        writer.write(molecule=bb1, path='bb1.xyz')

    .. testcode:: writing-to-a-file
        :hide:

        import os

        assert os.path.exists('bb1.xyz')

    .. testcleanup:: writing-to-a-file

        os.remove('bb1.xyz')

    """

    def _write_content(self, molecule, atom_ids):

        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )

        coords = molecule.get_position_matrix()
        content = [0]
        for i, atom_id in enumerate(atom_ids, 1):
            x, y, z = (i for i in coords[atom_id])
            atom, = molecule.get_atoms(atom_ids=atom_id)
            symbol = atom.__class__.__name__
            content.append(f'{symbol} {x:f} {y:f} {z:f}\n')
        # Set first line to the atom_count.
        content[0] = f'{i}\n\n'

        return content

    def to_string(self, molecule, atom_ids=None):
        """
        Get the ``.xyz`` string of  `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to `.xyz` format.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        Returns
        -------
        :class:`str`
            String in ``.xyz`` file format.

        """

        content = self._write_content(molecule, atom_ids)

        return ''.join(content)

    def write(self, molecule, path, atom_ids=None):
        """
        Write `molecule` to ``.xyz`` file format.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to ``.xyz`` format.

        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        Returns
        -------
        None : :class:`NoneType`
            A file is written.

        """

        content = self._write_content(molecule, atom_ids)

        with open(path, 'w') as f:
            f.write(''.join(content))



"""
Optimizer
=========

.. toctree::
    :maxdepth: 2

    Collapser <\
stk.molecular.topology_graphs.topology_graph.optimizers.collapser\
>

    Periodic Collapser <\
stk.molecular.topology_graphs.topology_graph.optimizers.\
periodic_collapser\
>

    MCHammer <\
stk.molecular.topology_graphs.topology_graph.optimizers.mchammer\
>
    Spinner <\
stk.molecular.topology_graphs.topology_graph.optimizers.spinner\
>

    NullOptimizer <\
stk.molecular.topology_graphs.topology_graph.optimizers.null\
>

"""


class Optimizer:
    """
    An abstract base class for optimizers.

    An optimizer is used to change the structure of the molecule under
    construction to be more realistic.

    """

    def optimize(self, state):
        """
        Optimize the structure of a molecule under construction.

        Parameters
        ----------
        state : :class:`.ConstructionState`
            The molecule being constructed.

        Returns
        -------
        :class:`.ConstructionState`
            The optimized construction state.

        """

        raise NotImplementedError()



"""
Linear
======

"""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from ....reactions import GenericReactionFactory
from ...topology_graph import (
    Edge,
    NullOptimizer,
    TopologyGraph,
    Vertex,
)
from .vertices import (
    HeadVertex,
    LinearVertex,
    TailVertex,
    UnaligningVertex,
)


class Linear(TopologyGraph):
    """
    Represents a linear polymer topology graph.

    Building blocks with two functional groups are required, unless the
    building block's position is specified to only be at the capping
    positions.

    Examples
    --------
    *Construction*

    Linear polymers require building blocks with two functional groups

    .. testcode:: construction

        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=4,
            ),
        )

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=4,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    polymer.get_atoms(),
                    polymer.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in polymer.get_bonds()
            ),
        )


    *Suggested Optimization*

    For :class:`.Linear` topologies, it is recommend to use the
    :class:`.Collapser` optimizer.

    .. testcode:: suggested-optimization

        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])

        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=4,
                # Setting scale_steps to False tends to lead to a
                # better structure.
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])

        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=4,
                # Setting scale_steps to False tends to lead to a
                # better structure.
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    polymer.get_atoms(),
                    polymer.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in polymer.get_bonds()
            ),
        )

    *Construction with Capping Units*

    Building blocks with a single functional group can
    also be provided as capping units

    .. testcode:: construction-with-capping-units

        import stk

        bb1 = stk.BuildingBlock('NCC(F)N', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb3 = stk.BuildingBlock('BrCCN', [stk.PrimaryAminoFactory()])

        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit='ABABC',
                num_repeating_units=1,
            ),
        )

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('NCC(F)N', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb3 = stk.BuildingBlock('BrCCN', [stk.PrimaryAminoFactory()])

        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit='ABABC',
                num_repeating_units=1,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    polymer.get_atoms(),
                    polymer.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in polymer.get_bonds()
            ),
        )

    *Defining the Orientation of Each Building Block*

    The `orientations` parameter allows the direction of each building
    block along to the chain to be flipped

    .. testcode:: defining-the-orientation-of-each-building-block

        import stk

        bb1 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb2 = stk.BuildingBlock(
            smiles='NC(Br)CN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )

        p1 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(1, 0.5),
            ),
        )

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('O=CC(F)CC=O', [stk.AldehydeFactory()])
        bb2 = stk.BuildingBlock(
            smiles='NC(Br)CN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )

        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(1, 0.5),
                random_seed=1,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    polymer.get_atoms(),
                    polymer.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in polymer.get_bonds()
            ),
        )

    In the above example, ``bb1`` is guaranteed to be flipped,
    ``bb2`` has a 50% chance of being flipped, each time it is placed
    on a node.

    Note that whether a building block will be flipped or not
    is decided during the initialization of :class:`.Linear`

    .. testcode:: defining-the-orientation-of-each-building-block

        # chain will always construct the same polymer.
        chain = stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
        )
        # p2 and p3 are guaranteed to be the same as they used the same
        # topology graph.
        p2 = stk.ConstructedMolecule(chain)
        p3 = stk.ConstructedMolecule(chain)

        # chain2 may lead to a different polymer than chain, despite
        # being initialized with the same parameters.
        chain2 = stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45)
        )

        # p4 and p5 are guaranteed to be the same because they used
        # the same topology graph. However, they may be different to
        # p2 and p3.
        p4 = stk.ConstructedMolecule(chain2)
        p5 = stk.ConstructedMolecule(chain2)

    The `random_seed` parameter can be used to get reproducible results

    .. testcode:: defining-the-orientation-of-each-building-block

        # p6 and p7 are guaranteed to be the same, because chain3 and
        # chain4 used the same random seed.

        chain3 = stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
            random_seed=4,
        )
        p6 = stk.ConstructedMolecule(chain3)

        chain4 = stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
            random_seed=4,
        )
        p7 = stk.ConstructedMolecule(chain4)

    *Using Numbers to Define the Repeating Unit*

    The repeating unit can also be specified through the indices of
    the building blocks

    .. testcode:: using-numbers-to-define-the-repeating-unit

        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb3 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])

        # p1 and p2 are different ways to write the same thing.
        p1 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit='ACB',
                num_repeating_units=1,
            ),
        )
        p2 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit=(0, 2, 1),
                num_repeating_units=1,
            ),
        )

    """

    def __init__(
        self,
        building_blocks,
        repeating_unit,
        num_repeating_units,
        orientations=None,
        random_seed=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
        optimizer=NullOptimizer(),
    ):
        """
        Initialize a :class:`Linear` instance.

        Parameters
        ----------
        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the polymer.

        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            A string specifying the repeating unit of the polymer.
            For example, ``'AB'`` or ``'ABB'``. The first building
            block passed to `building_blocks` is ``'A'`` and so on.

            The repeating unit can also be specified by the indices of
            `building_blocks`, for example ``'ABB'`` can be
            written as ``(0, 1, 1)``.

        num_repeating_units : :class:`int`
            The number of repeating units which are used to make the
            polymer.

        orientations : :class:`tuple` of :class:`float`, optional
            For each character in the repeating unit, a value
            between ``0`` and ``1`` (both inclusive) must be given in
            a :class:`tuple`. It indicates the probability that each
            monomer will have its orientation along the chain flipped.
            If ``0`` then the monomer is guaranteed not to flip. If
            ``1`` it is guaranteed to flip. This allows the user to
            create head-to-head or head-to-tail chains, as well as
            chain with a preference for head-to-head or head-to-tail if
            a number between ``0`` and ``1`` is chosen. If ``None``
            then ``0`` is picked in all cases.

            It is also possible to supply an orientation for every
            vertex in the final topology graph. In this case, the
            length of `orientations` must be equal to
            ``len(repeating_unit)*num_repeating_units``.

            If there is only one building block in the constructed
            polymer i.e. the `repeating_unit` has a length of 1 and
            `num_repeating_units` is 1, the building block will not
            be re-oriented, even if you provide a value to
            `orientations`.

        random_seed : :class:`int`, optional
            The random seed to use when choosing random orientations.

        reaction_factory : :class:`.ReactionFactory`, optional
            The factory to use for creating reactions between
            functional groups of building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        optimizer : :class:`.Optimizer`, optional
            Used to optimize the structure of the constructed
            molecule.

        Raises
        ------
        :class:`ValueError`
            If the length of `orientations` is not equal in length to
            `repeating_unit` or to the total number of vertices.

        """

        if orientations is None:
            orientations = tuple(
                0. for i in range(len(repeating_unit))
            )

        if len(orientations) == len(repeating_unit):
            orientations = orientations*num_repeating_units

        polymer_length = len(repeating_unit)*num_repeating_units
        if len(orientations) != polymer_length:
            raise ValueError(
                'The length of orientations must match either '
                'the length of repeating_unit or the '
                'total number of vertices.'
            )

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        try:
            head, *body, tail = orientations
            vertices_and_edges = self._get_vertices_and_edges(
                head_orientation=head,
                body_orientations=body,
                tail_orientation=tail,
                random_seed=random_seed,
            )
            vertices = vertices_and_edges.vertices
            edges = vertices_and_edges.edges

        except ValueError:
            vertices = (
                UnaligningVertex(0, (0., 0., 0.), False),
            )
            edges = ()

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(int(v.get_flip()) for v in vertices)

        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                building_blocks=building_blocks,
                vertices=vertices,
            ),
            edges=edges,
            reaction_factory=reaction_factory,
            construction_stages=(),
            optimizer=optimizer,
            num_processes=num_processes,
        )

    @staticmethod
    def _get_vertices_and_edges(
        head_orientation,
        body_orientations,
        tail_orientation,
        random_seed,
    ):
        """
        Get the vertices and edges of the topology graph.

        Parameters
        ----------
        head_orientation : :class:`float`
            The probability that the head vertex will do flipping.

        body_orientations : :class:`iterable` of :class:`float`
            For each body vertex, the probability that it will do
            flipping.

        tail_orientation : :class:`float`
            The probability that the tail vertex will do flipping.

        random_seed : :class:`int`
            The random seed to use.

        Returns
        -------
        :class:`.VerticesAndEdges`
            The vertices and edges of the topology graph.

        """

        generator = np.random.RandomState(random_seed)

        choices = [True, False]
        vertices = [
            HeadVertex(
                id=0,
                position=np.array([0, 0, 0]),
                flip=generator.choice(
                    a=choices,
                    p=[head_orientation, 1-head_orientation],
                ),
            ),
        ]
        edges = []
        for i, p in enumerate(body_orientations, 1):
            flip = generator.choice(choices, p=[p, 1-p])
            vertices.append(
                LinearVertex(i, np.array([i, 0, 0]), flip),
            )
            edges.append(Edge(len(edges), vertices[i-1], vertices[i]))

        vertices.append(
            TailVertex(
                id=len(vertices),
                position=np.array([len(vertices), 0, 0]),
                flip=generator.choice(
                    a=choices,
                    p=[tail_orientation, 1-tail_orientation],
                ),
            ),
        )

        edges.append(Edge(len(edges), vertices[-2], vertices[-1]))

        return _VerticesAndEdges(
            vertices=tuple(vertices),
            edges=tuple(edges),
        )

    def clone(self):
        clone = super().clone()
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        clone._orientations = self._orientations
        return clone

    @staticmethod
    def _normalize_repeating_unit(repeating_unit):
        if isinstance(repeating_unit, tuple):
            return repeating_unit

        base = ord('A')
        return tuple(ord(letter)-base for letter in repeating_unit)

    def _get_building_block_vertices(
        self,
        building_blocks,
        vertices,
    ):
        polymer = self._repeating_unit*self._num_repeating_units
        building_block_vertices = {}
        for bb_index, vertex in zip(polymer, vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)

        building_block_vertices = self._with_unaligning_vertices(
            building_block_vertices=building_block_vertices,
        )
        return building_block_vertices

    @staticmethod
    def _with_unaligning_vertices(
        building_block_vertices,
    ):
        clone = {}
        for building_block, vertices in (
            building_block_vertices.items()
        ):
            # Building blocks with 1 placer, cannot be aligned and
            # must therefore use an UnaligningVertex.
            if building_block.get_num_placers() == 1:
                clone[building_block] = tuple(
                    UnaligningVertex(
                        id=vertex.get_id(),
                        position=vertex.get_position(),
                        flip=vertex.get_flip(),
                    ) for vertex in vertices
                )
            else:
                clone[building_block] = vertices

        return clone

    def _get_scale(
        self,
        building_block_vertices,
    ) -> float:
        return max(
            bb.get_maximum_diameter()
            for bb in building_block_vertices
        )

    def __repr__(self):
        return (
            f'polymer.Linear({self._repeating_unit!r}, '
            f'{self._num_repeating_units!r}, '
            f'{self._orientations!r})'
        )


@dataclass(frozen=True)
class _VerticesAndEdges:
    vertices: tuple[Vertex]
    edges: tuple[Edge]



"""
Molecule Record
===============

"""

from stk.molecular import ConstructedMolecule


class MoleculeRecord:
    """
    An abstract base class for molecular records used by the EA.

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. This is a default implementation provided
    purely for convenience. Subclasses can freely ignore or
    override this implementation.

    """

    def __init__(self, topology_graph):
        """
        Initialize a :class:`.MoleculeRecord` instance.

        Parameters
        ----------
        topology_graph : :class:`.TopologyGraph`
            The topology graph of a :class:`.ConstructedMolecule`.

        """

        self._molecule = ConstructedMolecule(topology_graph)
        self._topology_graph = topology_graph
        self._fitness_value = None
        self._normalized_fitness_value = None

    def get_molecule(self):
        """
        Get the molecule held by the record.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The molecule held by the record.

        """

        return self._molecule

    def get_topology_graph(self):
        """
        Get the topology graph of the molecule.

        Returns
        -------
        :class:`.TopologyGraph`
            The topology graph.

        """

        return self._topology_graph

    def get_fitness_value(self, normalized=True):
        """
        Get the fitness value of the molecule in the record.

        Parameters
        ----------
        normalized : :class:`bool`, optional
            Toggles the return of the normalized vs unnormalized
            fitness value. The unnormalized fitness value is
            guaranteed to be constant for the same molecule
            across generations, while the normalized one is allowed
            to change.

        Returns
        -------
        :class:`object`
            The fitness value.

        None : :class:`NoneType`
            If a fitness value has not been assigned to the record.

        """

        return (
            self._normalized_fitness_value
            if normalized
            else self._fitness_value
        )

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.MoleculeRecord`
            The clone. Has the same type as the original record.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._molecule = self._molecule
        clone._topology_graph = self._topology_graph
        clone._fitness_value = self._fitness_value
        clone._normalized_fitness_value = (
            self._normalized_fitness_value
        )
        return clone

    def with_fitness_value(self, fitness_value, normalized=True):
        """
        Return a clone holding a different fitness value.

        Parameters
        ----------
        fitness_value : :class:`object`
            The fitness value of the clone.

        normalized : :class:`bool`, optional
            Toggles if the normalized or unnormalized fitness value is
            being set. If ``False``, both the normalized and
            unnormalized fitness values with be set to `fitness_value`.

        Returns
        -------
        :class:`.MoleculeRecord`
            The clone. Has the same type as the original record.

        """

        return self.clone()._with_fitness_value(
            fitness_value=fitness_value,
            normalized=normalized,
        )

    def _with_fitness_value(self, fitness_value, normalized):
        if not normalized:
            self._fitness_value = fitness_value
        self._normalized_fitness_value = fitness_value
        return self
